import re
import shlex
import mimetypes
from shlex import quote
from collections import deque
from typing import List
from dataclasses import dataclass, field

import hh
import orjson
from .curly import CURLArgumentParser


REGEX_SINGLE_OPT = re.compile(r'-\w$')


@dataclass
class ConversionResult:
    httpie: str
    errors: List[str] = field(default_factory=list)


def join_previous_arg(cmds: List[str], name: str):
    prev_arg = cmds[-1]
    if REGEX_SINGLE_OPT.match(prev_arg):
        cmds[-1] += name
    else:
        cmds.append(f'-{name}')


def curl_to_httpie(cmd: str, long_option: bool = False) -> ConversionResult:
    cargs = shlex.split(cmd)
    if not cargs:
        return ConversionResult('')
    if cargs[0] == 'curl':
        cargs = cargs[1:]
        if not cargs:
            return ConversionResult('http')
    args = CURLArgumentParser().parse_args(cargs)
    cmds = deque(['http'])
    if args.verbose:
        cmds.append('--verbose' if long_option else '-v')
    if args.location:
        if long_option:
            cmds.append('--follow')
        else:
            join_previous_arg(cmds, 'F')
    if args.remote_name:
        if long_option:
            cmds.append('--download')
        else:
            join_previous_arg(cmds, 'd')
    if args._data and not args._request_json:
        if long_option:
            cmds.append('--form')
        else:
            join_previous_arg(cmds, 'f')
    if args.user:
        if long_option:
            cmds.extend(('--auth', args.user))
        else:
            join_previous_arg(cmds, 'a')
            cmds.append(args.user)
    if args.include:
        cmds.append('--all')
    if args.insecure:
        cmds.append('--verify', 'no')
    if args.max_redirs:
        cmds.extend('--max-redirects', args.max_redirs)
    if args.head:
        cmds.append('HEAD')
    elif args.request and not (args._data and args.request == 'POST'):
        cmds.append(args.request)
    # URL
    cmds.append(args._url)
    # Headers
    for k, v in args._headers.items():
        cmds.append(f'{quote(k)}:{quote(v)}')
    if args._request_json and not args._data:
        mime = mimetypes.types_map['.json']
        cmds.append(f'{quote(hh.CONTENT_TYPE)}:{quote(mime)}')
    # Params
    for k, v in args._params:
        if k.startswith('-'):
            cmds.append('--')
        k = k.replace('=', r'\=')
        cmds.append(f'{quote(k)}=={quote(v)}')
    # Data
    for p, v in args._data:
        p = str(p)
        if p.startswith('-'):
            cmds.append('--')
        p = p.replace('=', r'\=')
        qp = quote(p)
        # Syntax for uploading file
        if isinstance(v, str) and v.startswith('@') and not args._request_json:
            cmds.append(f'{qp}@{quote(v)}')
            continue
        # Not uploading file
        # Python shlex's quote will turn bool value to empty string, that is not we want
        if isinstance(v, bool):
            js_bool = str(v).lower()
            cmds.append(f'{qp}:={js_bool}' if not args.get else f'{qp}=={str(v)}')
            continue
        try:
            qv = quote(v)
            cmds.append(f'{qp}={qv}' if not args.get else f'{qp}=={qv}')
        except TypeError:     # v is not string, normally after parsed from JSON
            if isinstance(v, (list, dict)):
                v = quote(orjson.dumps(v).decode())
            cmds.append(f'{qp}:={v}' if not args.get else f'{qp}=={quote(str(v))}')
    if args.output:
        param = '-o' if not long_option else '--output'
        cmds.extend((param, quote(args.output)))
    return ConversionResult(' '.join(cmds), args._errors)



import re
import sys
from pathlib import Path

from first import first


_REGEX_VERSION = re.compile(r'\s*version\s*=\s*["\']([.0-9a-z]+)["\']\s*$')

pyver = sys.version_info[:2]
if pyver <= (3, 7):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


def get_version(package: str, looked_path: Path):
    # Poetry don't include pyproject.toml file to distribution package,
    # so, if we see that file, its mean that the package is imported from development folder
    filepath = looked_path / 'pyproject.toml'   # type: Path
    if filepath.exists():
        found = first(_REGEX_VERSION.match(line) for line in filepath.open())
        if not found:
            return '0.0'
        return found.group(1)
    try:
        return importlib_metadata.version(package)
    except importlib_metadata.PackageNotFoundError:
        pass
    return '0.0'




import collections.abc
from typing import List, Optional, Tuple
from dataclasses import dataclass, field
from collections import OrderedDict, deque
from urllib.parse import parse_qsl
from json.decoder import JSONDecodeError

import hh
import yarl
import orjson
from tap import Tap

from .structures import CaseInsensitiveDict


@dataclass
class DataArgParseResult:
    data: List[Tuple[str, str]] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)


# Ref: https://helpmanual.io/help/curl/
class CURLArgumentParser(Tap):
    url: str
    verbose: bool = False
    include: bool = False
    location: bool = False
    insecure: bool = False
    remote_name: bool = False
    max_redirs: int = 0
    request: Optional[str] = None
    user: Optional[str] = None
    header: List[str] = []
    form: List[str] = []
    data: List[str] = []
    data_raw: List[str] = []
    head: bool = False
    get: bool = False
    output: Optional[str] = None
    http2: bool = False
    # Intermediate converted data
    _url: str = ''
    _params: List[Tuple[str, str]] = []
    _data: List[Tuple[str, str]] = []
    _headers: Optional[CaseInsensitiveDict] = None
    _request_json: bool = False
    _errors: List[str] = []

    def _get_class_variables(self) -> OrderedDict:
        '''Overide to exclude our private variables'''
        all_variables = super()._get_class_variables()
        return OrderedDict((k, v) for k, v in all_variables.items() if not k.startswith('_'))

    def add_arguments(self):
        self.add_argument('url')
        self.add_argument('-v', '--verbose')
        self.add_argument('-i', '--include')
        self.add_argument('-L', '--location')
        self.add_argument('-k', '--insecure')
        self.add_argument('-O', '--remote-name')
        self.add_argument('-X', '--request')
        self.add_argument('-u', '--user')
        self.add_argument('-H', '--header', nargs='?', action='append')
        self.add_argument('-d', '--data', nargs='?', action='append')
        self.add_argument('--data-raw', nargs='?', action='append')
        self.add_argument('-F', '--form')
        self.add_argument('-I', '--head')
        self.add_argument('-G', '--get')
        self.add_argument('-o', '--output')
        self._headers = CaseInsensitiveDict()

    def process_args(self):
        u = yarl.URL(self.url)
        # Clean fragment, if exist
        url = str(u.with_fragment(None))
        # Strip leading "http://" to be short
        self._url = url[7:] if u.scheme == 'http' else url
        self._params = u.query
        self._data = deque()
        for dstring in self.data:
            result = parse_post_data(dstring)
            self._data.extend(result.data)
            self._errors.extend(result.errors)
        for dstring in self.data_raw:
            result = parse_post_data(dstring, ignore_at=True)
            self._data.extend(result.data)
            self._errors.extend(result.errors)
        for h in self.header:
            k, v = h.split(':')
            k = k.strip()     # type: str
            v = v.strip()     # type: str
            if k.lower() == hh.CONTENT_TYPE.lower() and v.lower().endswith('/json'):
                self._request_json = True
                continue
            self._headers[k] = v

    def error(self, message):
        # Override to not let it exit our program
        pass


def parse_post_data(string: str, ignore_at: bool = False) -> DataArgParseResult:
    # https://ec.haxx.se/http/http-post
    if not string:
        return DataArgParseResult()
    data = parse_qsl(string)
    if data:
        return DataArgParseResult(data=data)
    # Standard parse_qsl failed to parse it
    if not ignore_at and '@' in string and not string.startswith('@'):
        # cURL spec says that the filename should already be url-encoded.
        key, filename = string.split('@')[:2]
        return DataArgParseResult(data=[(key, filename)])
    # HTTPie doesn't support sending raw content as request body
    # (though it allows to specify raw content as the value for a field),
    # so we can ignore cURL "content", "=content", "@filename" syntaxes.
    errors = deque()
    if string.startswith('@'):
        errors.append('@filename syntax (without field name) is not supported')
        return DataArgParseResult(data, errors)
    if string.startswith('='):
        errors.append('=content syntax (without field name) is not supported')
        return DataArgParseResult(data, errors)
    # Maybe JSON?
    try:
        jsdata = orjson.loads(string)
    except JSONDecodeError:
        # Not JSON
        errors.append('Cannot guess post data format')
        return DataArgParseResult(data, errors)
    if isinstance(jsdata, collections.abc.Mapping):
        data = list(jsdata.items())
        return DataArgParseResult(data, errors)
    errors.append('JSON content does not represent an object')
    return DataArgParseResult(errors=errors)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['curlipie']

package_data = \
{'': ['*']}

install_requires = \
['CaseInsensitiveDict>=1.0.0,<2.0.0',
 'first>=2.0.2,<3.0.0',
 'hh>=2.0.0,<3.0.0',
 'multidict>=4.7.4,<5.0.0',
 'orjson>=2.2.0,<3.0.0',
 'typed-argument-parser>=1.4,<2.0',
 'yarl>=1.4.2,<2.0.0']

setup_kwargs = {
    'name': 'curlipie',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Nguyễn Hồng Quân',
    'author_email': 'ng.hong.quan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
